@@ -1,8 +1,10 @@
 from collections import deque
+from functools import partial
 from itertools import (accumulate,
                        chain,
                        repeat)
 from typing import (Callable,
+                    Dict,
                     Iterable,
                     List,
                     Optional,
@@ -23,16 +25,14 @@ from reprit.base import generate_repr
 from sect.core.utils import (contour_to_edges_endpoints,
                              flatten,
                              pairwise)
-from .abcs import (QuadEdge,
-                   Triangulation)
 from .events_queue import EventsQueue
 from .hints import (QuaternaryPointPredicate,
                     SegmentEndpoints,
                     SegmentsRelater)
-from .quad_edge import (edge_to_endpoints,
+from .quad_edge import (QuadEdge,
+                        edge_to_endpoints,
                         edge_to_neighbours,
-                        edges_with_opposites,
-                        to_quad_edge_cls)
+                        edges_with_opposites)
 from .utils import (ceil_log2,
                     complete_vertices,
                     contour_to_oriented_edges_endpoints,
@@ -40,207 +40,177 @@ from .utils import (ceil_log2,
                     to_distinct)
 
 
-def triangulate_two_points(cls: Type[Triangulation],
-                           sorted_points: Sequence[Point]) -> Triangulation:
-    first_edge = cls.edge_cls.from_endpoints(*sorted_points)
-    return cls.from_sides(first_edge, first_edge.opposite)
+class Triangulation:
+    """Represents triangulation."""
 
+    @classmethod
+    def constrained_delaunay(cls,
+                             polygon: Polygon,
+                             *,
+                             extra_constraints: Sequence[Segment] = (),
+                             extra_points: Sequence[Point] = (),
+                             context: Context) -> 'Triangulation':
+        """
+        Constructs constrained Delaunay triangulation of given polygon
+        (with potentially extra points and constraints).
 
-def triangulate_three_points(cls: Type[Triangulation],
-                             sorted_points: Sequence[Point]) -> Triangulation:
-    left_point, mid_point, right_point = sorted_points
-    first_edge, second_edge = (
-        cls.edge_cls.from_endpoints(left_point, mid_point),
-        cls.edge_cls.from_endpoints(mid_point, right_point))
-    first_edge.opposite.splice(second_edge)
-    orientation = first_edge.orientation_of(right_point)
-    if orientation is Orientation.COUNTERCLOCKWISE:
-        second_edge.connect(first_edge)
-        return cls.from_sides(first_edge, second_edge.opposite)
-    elif orientation is Orientation.CLOCKWISE:
-        third_edge = second_edge.connect(first_edge)
-        return cls.from_sides(third_edge.opposite, third_edge)
-    else:
-        return cls.from_sides(first_edge, second_edge.opposite)
+        Based on
 
+        * divide-and-conquer algorithm by L. Guibas & J. Stolfi
+          for generating Delaunay triangulation,
 
-TriangulationBaseConstructor = Callable[[Type[Triangulation], Sequence[Point]],
-                                        Triangulation]
+        * algorithm by S. W. Sloan for adding constraints to Delaunay
+          triangulation,
 
+        * clipping algorithm by F. Martinez et al. for deleting in-hole
+          triangles.
 
-def to_triangulation_cls(context: Context) -> Type[Triangulation]:
-    """
-    Constructs triangulation class by given context.
+        Time complexity:
+            ``O(vertices_count * log vertices_count)`` for convex polygons
+            without extra constraints,
+            ``O(vertices_count ** 2)`` otherwise,
+            where ``vertices_count = len(border) + sum(map(len, holes))\
+     + len(extra_points) + len(extra_constraints)``.
+        Memory complexity:
+            ``O(vertices_count)``,
+            where ``vertices_count = len(border) + sum(map(len, holes))\
+     + len(extra_points) + len(extra_constraints)``.
+        Reference:
+            http://www.sccg.sk/~samuelcik/dgs/quad_edge.pdf
+            https://www.newcastle.edu.au/__data/assets/pdf_file/0019/22519/23_A-fast-algortithm-for-generating-constrained-Delaunay-triangulations.pdf
+            https://doi.org/10.1016/j.advengsoft.2013.04.004
+            http://www4.ujaen.es/~fmartin/bool_op.html
 
-    >>> from ground.base import get_context
-    >>> context = get_context()
-    >>> Contour, Point, Polygon = (context.contour_cls, context.point_cls,
-    ...                            context.polygon_cls)
-    >>> Triangulation = to_triangulation_cls(context)
-    >>> dart = Polygon(Contour([Point(0, 0), Point(3, 0), Point(1, 1),
-    ...                         Point(0, 3)]), [])
-    >>> (Triangulation.delaunay(dart.border.vertices).triangles()
-    ...  == [Contour([Point(0, 0), Point(3, 0), Point(1, 1)]),
-    ...      Contour([Point(0, 0), Point(1, 1), Point(0, 3)]),
-    ...      Contour([Point(0, 3), Point(1, 1), Point(3, 0)])])
-    True
-    >>> (Triangulation.constrained_delaunay(dart).triangles()
-    ...  == [Contour([Point(0, 0), Point(3, 0), Point(1, 1)]),
-    ...      Contour([Point(0, 0), Point(1, 1), Point(0, 3)])])
-    True
-    """
+        :param polygon: target polygon.
+        :param extra_points:
+            additional points to be presented in the triangulation.
+        :param extra_constraints:
+            additional constraints to be presented in the triangulation.
+        :param context: geometric context.
+        :returns:
+            triangulation of the border, holes & extra points
+            considering constraints.
 
-    base_cases = {2: triangulate_two_points,
-                  3: triangulate_three_points}
+        >>> from ground.base import get_context
+        >>> context = get_context()
+        >>> Contour, Point, Polygon = (context.contour_cls, context.point_cls,
+        ...                            context.polygon_cls)
+        >>> dart = [Point(0, 0), Point(3, 0), Point(1, 1), Point(0, 3)]
+        >>> (Triangulation.constrained_delaunay(Polygon(Contour(dart), []),
+        ...                                     context=context).triangles()
+        ...  == [Contour([Point(0, 0), Point(3, 0), Point(1, 1)]),
+        ...      Contour([Point(0, 0), Point(1, 1), Point(0, 3)])])
+        True
+        """
+        border, holes = polygon.border, polygon.holes
+        if extra_points:
+            border, holes, extra_points = complete_vertices(
+                    border, holes, extra_points, context)
+        result = cls.delaunay(list(chain(border.vertices,
+                                         flatten(hole.vertices
+                                                 for hole in holes),
+                                         extra_points)),
+                              context=context)
+        border_endpoints = contour_to_edges_endpoints(border)
+        constrain(result, chain(border_endpoints,
+                                flatten(map(contour_to_edges_endpoints,
+                                            holes)),
+                                [(segment.start, segment.end)
+                                 for segment in extra_constraints]))
+        bound(result, border_endpoints)
+        cut(result, holes)
+        result._triangular_holes_vertices.update(
+                frozenset(hole.vertices)
+                for hole in holes
+                if len(hole.vertices) == 3)
+        return result
 
-    class Result(Triangulation):
-        __slots__ = 'left_side', 'right_side', '_triangular_holes_vertices'
+    @classmethod
+    def delaunay(cls,
+                 points: Sequence[Point],
+                 *,
+                 context: Context) -> 'Triangulation':
+        """
+        Constructs Delaunay triangulation of given points.
 
-        edge_cls = to_quad_edge_cls(context)
+        Based on divide-and-conquer algorithm by L. Guibas & J. Stolfi.
 
-        def __init__(self, left_side: QuadEdge, right_side: QuadEdge) -> None:
-            self.left_side, self.right_side = left_side, right_side
-            self._triangular_holes_vertices = set()
+        Time complexity:
+            ``O(len(points) * log len(points))``
+        Memory complexity:
+            ``O(len(points))``
+        Reference:
+            http://www.sccg.sk/~samuelcik/dgs/quad_edge.pdf
 
-        __repr__ = generate_repr(__init__)
+        :param points: 3 or more points to triangulate.
+        :param context: geometric context.
+        :returns: triangulation of the points.
 
-        @classmethod
-        def constrained_delaunay(cls,
-                                 polygon: Polygon,
-                                 *,
-                                 extra_constraints: Sequence[Segment] = (),
-                                 extra_points: Sequence[Point] = ()
-                                 ) -> Triangulation:
-            """
-            Constructs constrained Delaunay triangulation of given polygon
-            (with potentially extra points and constraints).
+        >>> from ground.base import get_context
+        >>> context = get_context()
+        >>> Contour, Point = context.contour_cls, context.point_cls
+        >>> dart = [Point(0, 0), Point(3, 0), Point(1, 1), Point(0, 3)]
+        >>> (Triangulation.delaunay(dart,
+        ...                         context=context).triangles()
+        ...  == [Contour([Point(0, 0), Point(3, 0), Point(1, 1)]),
+        ...      Contour([Point(0, 0), Point(1, 1), Point(0, 3)]),
+        ...      Contour([Point(0, 3), Point(1, 1), Point(3, 0)])])
+        True
+        """
+        points = sorted(to_distinct(points))
+        lengths = coin_change(len(points), base_cases)
+        result = [cls._initialize_triangulation(points[start:stop],
+                                                context)
+                  for start, stop in pairwise(accumulate((0,) + lengths))]
+        for _ in repeat(None, ceil_log2(len(result))):
+            parts_to_merge_count = len(result) // 2 * 2
+            result = ([merge(result[offset], result[offset + 1])
+                       for offset in range(0, parts_to_merge_count, 2)]
+                      + result[parts_to_merge_count:])
+        return result[0]
 
-            Based on
+    __slots__ = ('context', 'left_side', 'right_side',
+                 '_triangular_holes_vertices')
 
-            * divide-and-conquer algorithm by L. Guibas & J. Stolfi
-              for generating Delaunay triangulation,
+    def __init__(self,
+                 left_side: QuadEdge,
+                 right_side: QuadEdge,
+                 context: Context) -> None:
+        self.context, self.left_side, self.right_side = (context, left_side,
+                                                         right_side)
+        self._triangular_holes_vertices = set()
 
-            * algorithm by S. W. Sloan for adding constraints to Delaunay
-              triangulation,
+    __repr__ = generate_repr(__init__)
 
-            * clipping algorithm by F. Martinez et al. for deleting in-hole
-              triangles.
+    def delete(self, edge: QuadEdge) -> None:
+        """Deletes given edge from the triangulation."""
+        if edge is self.right_side or edge.opposite is self.right_side:
+            self.right_side = self.right_side.right_from_end.opposite
+        if edge is self.left_side or edge.opposite is self.left_side:
+            self.left_side = self.left_side.left_from_start
+        edge.delete()
 
-            Time complexity:
-                ``O(vertices_count * log vertices_count)`` for convex polygons
-                without extra constraints,
-                ``O(vertices_count ** 2)`` otherwise,
-                where ``vertices_count = len(border) + sum(map(len, holes))\
-         + len(extra_points) + len(extra_constraints)``.
-            Memory complexity:
-                ``O(vertices_count)``,
-                where ``vertices_count = len(border) + sum(map(len, holes))\
-         + len(extra_points) + len(extra_constraints)``.
-            Reference:
-                http://www.sccg.sk/~samuelcik/dgs/quad_edge.pdf
-                https://www.newcastle.edu.au/__data/assets/pdf_file/0019/22519/23_A-fast-algortithm-for-generating-constrained-Delaunay-triangulations.pdf
-                https://doi.org/10.1016/j.advengsoft.2013.04.004
-                http://www4.ujaen.es/~fmartin/bool_op.html
+    def triangles(self) -> List[Contour]:
+        """Returns triangles of the triangulation."""
+        vertices_sets = to_distinct(
+                frozenset((edge.start, edge.end, edge.left_from_start.end))
+                for edge in to_edges(self)
+                if (edge.left_from_start.end
+                    == edge.opposite.right_from_start.end
+                    and (edge.orientation_of(edge.left_from_start.end)
+                         is Orientation.COUNTERCLOCKWISE)))
+        contour_cls, orienteer = (self.context.contour_cls,
+                                  self.context.angle_orientation)
+        return [contour_cls(normalize_contour_vertices(list(vertices),
+                                                       orienteer))
+                for vertices in vertices_sets
+                if vertices not in self._triangular_holes_vertices]
 
-            :param polygon: target polygon.
-            :param extra_points:
-                additional points to be presented in the triangulation.
-            :param extra_constraints:
-                additional constraints to be presented in the triangulation.
-            :returns:
-                triangulation of the border, holes & extra points
-                considering constraints.
-            """
-            border, holes = polygon.border, polygon.holes
-            if extra_points:
-                border, holes, extra_points = complete_vertices(
-                        border, holes, extra_points,
-                        context=context)
-            result = cls.delaunay(list(chain(border.vertices,
-                                             flatten(hole.vertices
-                                                     for hole in holes),
-                                             extra_points)))
-            border_endpoints = contour_to_edges_endpoints(border)
-            constrain(result, chain(border_endpoints,
-                                    flatten(map(contour_to_edges_endpoints,
-                                                holes)),
-                                    [(segment.start, segment.end)
-                                     for segment in extra_constraints]),
-                      context=context)
-            bound(result, border_endpoints)
-            cut(result, holes,
-                context=context)
-            result._triangular_holes_vertices.update(
-                    frozenset(hole.vertices)
-                    for hole in holes
-                    if len(hole.vertices) == 3)
-            return result
-
-        @classmethod
-        def delaunay(cls, points: Sequence[Point]) -> 'Result':
-            """
-            Constructs Delaunay triangulation of given points.
-
-            Based on divide-and-conquer algorithm by L. Guibas & J. Stolfi.
-
-            Time complexity:
-                ``O(len(points) * log len(points))``
-            Memory complexity:
-                ``O(len(points))``
-            Reference:
-                http://www.sccg.sk/~samuelcik/dgs/quad_edge.pdf
-
-            :param points: 3 or more points to triangulate.
-            :returns: triangulation of the points.
-            """
-            points = sorted(to_distinct(points))
-            lengths = coin_change(len(points), base_cases)
-            result = [cls._initialize_triangulation(points[start:stop])
-                      for start, stop in pairwise(accumulate((0,) + lengths))]
-            incircle_test = context.point_point_point_incircle_test
-            for _ in repeat(None, ceil_log2(len(result))):
-                parts_to_merge_count = len(result) // 2 * 2
-                result = ([merge(result[offset], result[offset + 1],
-                                 incircle_test)
-                           for offset in range(0, parts_to_merge_count, 2)]
-                          + result[parts_to_merge_count:])
-            return result[0]
-
-        @classmethod
-        def from_sides(cls,
-                       left_side: QuadEdge,
-                       right_side: QuadEdge) -> Triangulation:
-            return cls(left_side, right_side)
-
-        def delete(self, edge: QuadEdge) -> None:
-            if edge is self.right_side or edge.opposite is self.right_side:
-                self.right_side = self.right_side.right_from_end.opposite
-            if edge is self.left_side or edge.opposite is self.left_side:
-                self.left_side = self.left_side.left_from_start
-            edge.delete()
-
-        def triangles(self) -> List[Contour]:
-            vertices_sets = to_distinct(
-                    frozenset((edge.start, edge.end, edge.left_from_start.end))
-                    for edge in to_edges(self)
-                    if (edge.left_from_start.end
-                        == edge.opposite.right_from_start.end
-                        and (edge.orientation_of(edge.left_from_start.end)
-                             is Orientation.COUNTERCLOCKWISE)))
-            contour_cls, orienteer = (context.contour_cls,
-                                      context.angle_orientation)
-            return [contour_cls(normalize_contour_vertices(list(vertices),
-                                                           orienteer))
-                    for vertices in vertices_sets
-                    if vertices not in self._triangular_holes_vertices]
-
-        @classmethod
-        def _initialize_triangulation(cls, points: Sequence[Point]
-                                      ) -> Triangulation:
-            return base_cases[len(points)](cls, points)
-
-    Result.__name__ = Result.__qualname__ = Triangulation.__name__
-    return Result
+    @classmethod
+    def _initialize_triangulation(cls,
+                                  points: Sequence[Point],
+                                  context: Context) -> 'Triangulation':
+        return base_cases[len(points)](cls, points, context)
 
 
 def bound(triangulation: Triangulation,
@@ -278,13 +248,12 @@ def connect(base_edge: QuadEdge,
 
 
 def constrain(triangulation: Triangulation,
-              constraints: Iterable[SegmentEndpoints],
-              *,
-              context: Context) -> None:
+              constraints: Iterable[SegmentEndpoints]) -> None:
     endpoints = {edge_to_endpoints(edge) for edge in to_edges(triangulation)}
     inner_edges = to_unique_inner_edges(triangulation)
-    incircle_test, segments_relater = (context.point_point_point_incircle_test,
-                                       context.segments_relation)
+    incircle_test, segments_relater = (
+        triangulation.context.point_point_point_incircle_test,
+        triangulation.context.segments_relation)
     for constraint in constraints:
         constraint_endpoints = frozenset(constraint)
         if constraint_endpoints in endpoints:
@@ -305,22 +274,20 @@ def constrain(triangulation: Triangulation,
         inner_edges.update(new_edges)
 
 
-def cut(triangulation: Triangulation,
-        holes: Sequence[Contour],
-        *,
-        context: Context) -> None:
+def cut(triangulation: Triangulation, holes: Sequence[Contour]) -> None:
     if not holes:
         return
-    events_queue = EventsQueue(context)
+    events_queue = EventsQueue(triangulation.context)
     for edge in to_unique_inner_edges(triangulation):
         events_queue.register_edge(edge,
                                    from_left=True,
                                    is_counterclockwise_contour=True)
+    orienteer = triangulation.context.angle_orientation
     for hole in holes:
         for endpoints in contour_to_oriented_edges_endpoints(
                 hole,
                 clockwise=True,
-                orienteer=context.angle_orientation):
+                orienteer=orienteer):
             events_queue.register_segment(endpoints,
                                           from_left=False,
                                           is_counterclockwise_contour=False)
@@ -378,11 +345,10 @@ def is_convex_quadrilateral_diagonal(edge: QuadEdge) -> bool:
                     edge.right_from_start.end))
 
 
-def merge(left: Triangulation,
-          right: Triangulation,
-          incircle_test: QuaternaryPointPredicate) -> Triangulation:
-    connect(find_base_edge(left, right), incircle_test)
-    return left.from_sides(left.left_side, right.right_side)
+def merge(left: Triangulation, right: Triangulation) -> Triangulation:
+    connect(find_base_edge(left, right),
+            left.context.point_point_point_incircle_test)
+    return type(left)(left.left_side, right.right_side, left.context)
 
 
 def resolve_crossings(crossings: List[QuadEdge],
@@ -489,3 +455,39 @@ def to_unique_edges(triangulation: Triangulation) -> Iterable[QuadEdge]:
 def to_unique_inner_edges(triangulation: Triangulation) -> Set[QuadEdge]:
     return (set(to_unique_edges(triangulation))
             .difference(to_boundary_edges(triangulation)))
+
+
+BaseCase = Callable[[Type[Triangulation], Sequence[Point], Context],
+                    Triangulation]
+base_cases = {}  # type: Dict[int, BaseCase]
+register_base_case = partial(partial, base_cases.setdefault)
+
+
+@register_base_case(2)
+def triangulate_two_points(cls: Type[Triangulation],
+                           sorted_points: Sequence[Point],
+                           context: Context) -> Triangulation:
+    first_edge = QuadEdge.from_endpoints(*sorted_points,
+                                         context=context)
+    return cls(first_edge, first_edge.opposite, context)
+
+
+@register_base_case(3)
+def triangulate_three_points(cls: Type[Triangulation],
+                             sorted_points: Sequence[Point],
+                             context: Context) -> Triangulation:
+    left_point, mid_point, right_point = sorted_points
+    first_edge, second_edge = (QuadEdge.from_endpoints(left_point, mid_point,
+                                                       context=context),
+                               QuadEdge.from_endpoints(mid_point, right_point,
+                                                       context=context))
+    first_edge.opposite.splice(second_edge)
+    orientation = first_edge.orientation_of(right_point)
+    if orientation is Orientation.COUNTERCLOCKWISE:
+        second_edge.connect(first_edge)
+        return cls(first_edge, second_edge.opposite, context)
+    elif orientation is Orientation.CLOCKWISE:
+        third_edge = second_edge.connect(first_edge)
+        return cls(third_edge.opposite, third_edge, context)
+    else:
+        return cls(first_edge, second_edge.opposite, context)
